# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import unittest

from universa.types import PrivateKey, PublicKey, KeyAddress


class TestCrypto(unittest.TestCase):
    def test(self):
        private = PrivateKey(size=2048)
        private_2 = PrivateKey(key=private.key)
        self.assertTrue(private.equals(private_2), 'equals for private keys failed')
        self.assertEqual(private.key, private_2.key, 'private keys are different')

        public = private.public_key
        public_3 = PublicKey(key=private.public_key.key)
        self.assertTrue(public_3.equals(public), 'equals for public keys failed')
        self.assertEqual(public_3.key, public.key, 'public keys are different')

        key_address = public.short_address
        self.assertFalse(key_address.is_long, 'short address should be `not is_long`')

        key_address_4 = KeyAddress(address=key_address.address)
        self.assertTrue(key_address_4.equals(key_address), 'equals for key addresses failed')
        self.assertTrue(key_address_4.is_matching_key_address(key_address), 'is_matching_key_address failed')
        self.assertEqual(key_address_4.address, key_address.address, 'key addresses are different')
        self.assertEqual(key_address_4.uaddress, key_address.uaddress, 'key addresses are different')

        key_address_5 = KeyAddress(uaddress=key_address.uaddress)
        self.assertTrue(key_address_5.equals(key_address), 'equals for key addresses failed')

        long_address = public.long_address
        self.assertTrue(long_address.is_long, 'long address should be is_long')


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

from .base import Base


#
# Abstract
#
class Role(Base):
    class RequiredMode:
        ALL_OF = 'ALL_OF'  # In this mode, all references must be allowed for allows role
        ANY_OF = 'ANY_OF'  # In this mode, at least one of the references must be allowed for allows role
        MODES = [ALL_OF, ANY_OF]

    def __init__(self, name, **kwargs):
        self.name = name
        super(Role, self).__init__(name=name, **kwargs)

    def _instantiate_data(self):
        return self.name

    def anonymize(self):
        self._invoke('anonymize')

    def equals_ignore_name(self, obj):
        return self._equals('equalsIgnoreName', obj)

    def equals_ignore_name_and_refs(self, obj):
        return self._equals('equalsIgnoreNameAndRefs', obj)

    def get_comment(self):
        return self._invoke('getComment')

    def get_contract(self):
        from universa.types.contract import Contract
        rsp = self._invoke('getContract')
        return Contract.get(rsp['id']) if rsp is not None else None

    def get_name(self):
        return self._invoke('getName')

    def get_references(self, required_mode):
        if required_mode in [self.RequiredMode.MODES]:
            return self._invoke('getReferences', required_mode)
        return None

    def hash_code(self):
        return self._invoke('hashCode')

    def is_valid(self):
        return self._invoke('isValid')

    def link_as(self, role_name):
        rsp = self._invoke('linkAs', role_name)
        return RoleLink.get(rsp['id'])

    def resolve(self):
        """
        Goes through the links to the first non-link role and returns the certain Role.
        May return None if RoleLink target is None (or is a recursive link).

        :rtype: Role | None
        """
        rsp = self._invoke('resolve')
        if rsp is not None:
            klass = self.get_class_by_java(rsp['className'])
            return klass.get(rsp['id'])
        else:
            return None

    def set_comment(self, comment):
        self._invoke('setComment', comment)

    def set_contract(self, contract):
        """
        :type contract: universa.types.contract.Contract
        :return:
        """
        self._invoke('setContract', contract.remote)

    def to_string(self):
        return self._invoke('toString')


#
# Real
#
class SimpleRole(Role):
    JAVA_CLASS = 'com.icodici.universa.contract.roles.SimpleRole'

    def __init__(self, name, records=None, **kwargs):
        """
        :type name: str
        :type records: list[KeyAddress] | None
        """
        self.__records = records
        super(SimpleRole, self).__init__(name=name, **kwargs)
        del self.__records

    def _instantiate_data(self):
        if self.__records is not None:
            return self.name, [r.remote for r in self.__records]
        return self.name

    def get_simple_anonymous_ids(self):
        return self._get_hashset('getSimpleAnonymousIds')

    def get_simple_key_addresses(self):
        return self._get_hashset('getSimpleKeyAddresses')

    def get_simple_key_records(self):
        return self._get_hashset('getSimpleKeyRecords')

    def get_simple_keys(self):
        return self._get_hashset('getSimpleKeys')


class ListRole(Role):
    JAVA_CLASS = 'com.icodici.universa.contract.roles.ListRole'

    class Mode:
        ALL = 'ALL'  # Role could be performed only if set of keys could play all sub-roles
        ANY = 'ANY'  # Role could be performed if set of keys could play any role of the list
        QUORUM = 'QUORUM'  # Role could be played if set of keys could play any quorrum set of roles, e.g.
        MODES = [ALL, ANY, QUORUM]


class RoleLink(Role):
    JAVA_CLASS = 'com.icodici.universa.contract.roles.RoleLink'

    def __init__(self, name, target_name, **kwargs):
        self.target_name = target_name
        super(RoleLink, self).__init__(name=name, target_name=target_name, **kwargs)

    def get_role(self):
        """
        Returns the target role it links to (may return one more RoleLink)
        and .resolve() goes through all links to find the first non-link role.

        :rtype: Role
        """
        rsp = self._invoke('getRole')
        if rsp is not None:
            klass = self.get_class_by_java(rsp['className'])
            return klass.get(rsp['id'])
        else:
            return None



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals


class UniversaException(Exception):
    def __init__(self, message, error):
        super(UniversaException, self).__init__(message)
        self.error = error



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import unittest

from universa.types import PrivateKey, SimpleRole, HashSet


class TestHash(unittest.TestCase):
    def test(self):
        private_key = PrivateKey(size=2048)
        short_address = private_key.public_key.short_address
        role = SimpleRole('role', [short_address])
        x = role._invoke('getSimpleKeyAddresses')

        hashset = HashSet(id=x['id'])

        self.assertTrue(hashset[0].equals(short_address))


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import json
import os
import pexpect
import socket
import streamexpect
import weakref

from universa import logging
from universa import exceptions

logger = logging.getLogger()


class Transport(object):
    DEFAULT_CONNECTION_METHOD = 'pipe'
    DEFAULT_BINARY = 'umi'
    DEFAULT_CONNECTION_CONFIG = (DEFAULT_CONNECTION_METHOD, DEFAULT_BINARY)
    WINDOW = 1024 * 1024 * 5

    OBJECTS = {}

    method = DEFAULT_CONNECTION_METHOD
    __instance = None
    __proc = None

    def __init__(self, serial=0):
        if Transport.__instance is not None:
            raise Exception('Universa Transport is a singleton.')

        Transport.__instance = self
        self.OBJECTS = weakref.WeakValueDictionary()
        self.serial = max(serial, 0)

    def __del__(self):
        logger.debug('Cleaning')
        if self.__proc is not None:
            self.drop_objects(drop_all=True)
            self.__proc.close()

    @classmethod
    def setupUMI(cls, method, path=None, host=None, port=None, serial=0):
        """Configure Python to use a specific UMI connection.

        :param method: one of `'pipe'`, `'tcp'` or `'unix'`.
            If `path='pipe'`, you must provide the `path` argument with the path to the UMI executable;
            may be in $PATH.
            If `path='tcp'`, you must provide the `host` and `port` arguments of the UMI TCP server.
            If `path='unix'`, you must provide the `path` argument to connect to the UMI TCP server.
        :return: the transport (pexpect) instance
        :raises ValueError: if the arguments are improperly provided;
            if UMI executable is not available or reachable.
        """
        if method == 'pipe':
            if not path:
                raise ValueError('In "pipe" mode, you should provide non-empty "path" for UMI binary!')
            builder = lambda: cls.__make_pipe_transport(path)
        elif method == 'tcp':
            if not host or not port:
                raise ValueError('In "tcp" mode, you should provide non-empty "host" and "port" for UMI TCP socket!')
            builder = lambda: cls.__make_tcp_transport(host, port)
        elif method == 'unix':
            if not path:
                raise ValueError('In "unix" mode, you should provide non-empty "path" for UMI Unix socket!')
            builder = lambda: cls.__make_unix_transport(path)
        else:
            raise ValueError('Unsupported method {}'.format(method))

        # We have a `builder` defined here.
        # Do we have a running transport already?
        if cls.__proc is not None:
            try:
                cls.__instance.drop_objects(drop_all=True)
                cls.__proc.close()
            except OSError:
                pass

        cls.__proc = builder()
        cls.method = method
        transport.serial = max(serial, 0)

    @staticmethod
    def __make_pipe_transport(binary_path):
        try:
            return pexpect.spawn(binary_path, timeout=None, maxread=Transport.WINDOW)
        except pexpect.exceptions.ExceptionPexpect as e:
            if e.value.startswith('The command was not found or was not executable:'):
                raise ValueError('UMI executable you have chosen is not available at the requested path (or in $PATH).')
            else:
                raise

    @staticmethod
    def __make_tcp_transport(host, port):
        try:
            sock = socket.create_connection((host, port))
        except ConnectionRefusedError:
            raise ConnectionRefusedError('UMI refused the connection at tcp://%s:%s.' % (host, port))
        return streamexpect.wrap(sock, window=Transport.WINDOW, close_stream=False)

    @staticmethod
    def __make_unix_transport(socket_path):
        socket_path = os.path.abspath(socket_path)
        try:
            sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            sock.connect(socket_path)
        except ConnectionRefusedError:
            raise ConnectionRefusedError('UMI refused the connection at unix://%s.' % socket_path)
        return streamexpect.wrap(sock, window=Transport.WINDOW, close_stream=False)

    @staticmethod
    def get_instance():
        if Transport.__instance is None:
            Transport()
        return Transport.__instance

    @property
    def transport(self):
        if self.__proc is None:
            self.setupUMI(*self.DEFAULT_CONNECTION_CONFIG)  # default UMI configuration if not set up explicitly
        return self.__proc

    @staticmethod
    def _format(**kwargs):
        return json.dumps(kwargs, separators=(',', ':')).encode()

    def sync_call(self, name, full_response=False, **kwargs):
        cmd = self._format(serial=self.serial, cmd=name, **kwargs)
        logger.debug('   >> %s', cmd.decode())
        if self.method == 'pipe':
            self.transport.sendline(cmd)
            self.transport.expect('\r\n.*"ref":%s.*\r\n' % self.serial)
            rsp_text = self.transport.after
        else:
            self.transport.sendall(cmd + b'\r\n')
            match = self.transport.expect_regex((r'({.*"ref":%s.*})' % self.serial).encode())
            rsp_text = match.groups[0]

        try:
            rsp_text = rsp_text.decode()
        except Exception:
            pass
        finally:
            rsp_text = rsp_text.strip()

        self.serial += 1
        logger.debug('   << %s', rsp_text)
        rsp = json.loads(rsp_text)
        if 'error' in rsp:
            raise exceptions.UniversaException(rsp_text, rsp['error'])

        if not full_response:
            return rsp['result']
        else:
            return rsp

    @property
    def version(self):
        return self.sync_call('version')

    def test(self):
        return self.sync_call('version', full_response=True)

    def instantiate(self, object_type, *args):
        return self.sync_call('instantiate', args=[object_type] + list(args))

    def invoke(self, remote_object_id, method_name, *args):
        return self.sync_call('invoke', args=[remote_object_id, method_name] + list(args))

    def invoke_static(self, object_type, method_name, *args):
        return self.sync_call('invoke', args=[object_type, method_name] + list(args))

    def get(self, remote_object_id):
        return self.sync_call('get', args=[remote_object_id])

    def get_field(self, remote_object_id, field_name):
        return self.sync_call('get_field', args=[remote_object_id, field_name])

    def drop_objects(self, ids=None, drop_all=False):
        if not drop_all:
            to_delete = set()
            for _id in ids or []:
                to_delete.add(_id)
                if _id in self.OBJECTS:
                    self.OBJECTS.pop(_id, None)
            else:
                return self.sync_call('drop_objects', ids=list(to_delete))
        else:
            self.OBJECTS = weakref.WeakValueDictionary()
            return self.sync_call('drop_all')


transport = Transport()

setupUMI = Transport.setupUMI



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

try:
    from collections.abc import MutableMapping  # noqa
except ImportError:
    from collections import MutableMapping  # noqa
import json

from .base import Base


class Binder(Base, MutableMapping):
    JAVA_CLASS = 'net.sergeych.tools.Binder'

    LOCAL_ATTRS = ('id', '__initial')

    def __init__(self, *args, **kwargs):
        _id = kwargs.pop('id', None)
        self.__initial = dict(*args, **kwargs)
        super(Binder, self).__init__(id=_id)

    @property
    def __dict__(self):
        return self._get(self.id)

    def __getitem__(self, key):
        item = self._invoke('get', key)
        if isinstance(item, dict) and item.get('__type') == 'RemoteObject':
            remote_id = item.get('id')
            remote_class = item.get('className')
            if remote_id is not None and remote_class:
                return self.get_class_by_java(remote_class)(id=remote_id)
        return item

    def __setitem__(self, key, value):
        self._invoke('set', key, value)

    def __delitem__(self, key):
        self._invoke('remove', key)

    def __iter__(self):
        return iter(self.__dict__.items())

    def __len__(self):
        return len(self.__dict__)

    def setdefault(self, k, default=None):
        self._invoke('putIfAbsent', k, self.obj_umification(default))
        return self[k]

    def _instantiate_data(self):
        return self.kwargs_umification(self.__initial)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

from .base import Base
from .hash_id import HashId


class Client(Base):
    JAVA_CLASS = 'com.icodici.universa.node2.network.Client'

    def __init__(self, topology_input, topology_cache_dir, client_private_key, **kwargs):
        self.topology_input = topology_input
        self.topology_cache_dir = topology_cache_dir
        self.client_private_key = client_private_key
        super(Client, self).__init__()

    def _instantiate_data(self):
        return (self.topology_input,
                self.topology_cache_dir,
                self.client_private_key.remote
                    if isinstance(self.client_private_key, Base)
                    else self.client_private_key)

    def get_version(self):
        return self._invoke('getVersion')

    def get_state(self, hash_id):
        """
        :type hash_id: HashId
        :rtype: dict
        """
        return self._invoke('getState', hash_id)

    def register_parcel_with_state(self, packed, millis_to_wait):
        """
        Register the contract on the network using parcel (to provide payment)
        that waits until registration is complete for a given amount of time.

        :param bytes packed: Parcel binary
        :param int millis_to_wait: maximum time to wait for final ItemState
        :rtype: dict
        """
        return self._invoke('registerParcelWithState', packed, millis_to_wait)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

from datetime import timedelta
import unittest
import warnings

from universa.types import PrivateKey, Contract, SimpleRole


class TestContract(unittest.TestCase):
    def test_contracts(self):
        private_key = PrivateKey(size=2048)

        short_address = private_key.public_key.short_address

        sr = SimpleRole('name', [short_address])

        contract = Contract()

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter('always')
            contract.set_creator_keys(private_key)
            contract.set_issuer_keys(private_key.public_key)

            assert len(w) == 4
            assert issubclass(w[0].category, PendingDeprecationWarning)  # .set_creator_keys will be deprecated
            assert issubclass(w[1].category, SyntaxWarning)  # You shouldn't use PrivateKey for roles! Never!
            assert issubclass(w[2].category, PendingDeprecationWarning)  # .set_issuer_keys will be deprecated
            assert issubclass(w[3].category, PendingDeprecationWarning)  # Use KeyAddress instead of PublicKey

        contract.set_owner_addresses(private_key.public_key.short_address)
        contract.set_owner_addresses(private_key.public_key.short_address.address)
        contract.set_owner_addresses(private_key.public_key.short_address.uaddress)

        assert contract.expires_at is None, 'expires_at should be None'
        contract.expires_at = contract.created_at + timedelta(days=90)
        assert (contract.expires_at - contract.created_at).days == 90, 'expires_at should be None'

        sealed = contract.seal()

        # contract_2 = Contract(binary=sealed)
        # assert contract_2.equals(contract), 'contracts should be equal'
        #
        #
        # simple_role = contract_2.get_owner()
        #
        # contract_3 = simple_role.get_contract()
        # assert contract_3.equals(contract), 'contracts should be equal'

    def test_roles(self):
        simple_role = SimpleRole('my role')
        assert simple_role.get_contract() is None, 'contract should be None for role without a contract'


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

from .base import DEFAULT_CLASS, TYPES_CLASS_MAP, JAVA_CLASS_MAP, Base
from .binder import Binder
from .contract import Contract, TransactionPack, ContractsService, Reference, Parcel, ExtendedSignature
from .crypto import KeyAddress, PrivateKey, PublicKey
from .hash_id import HashId
from .network import Client
from .permissions import ChangeNumberPermission, ChangeOwnerPermission, ModifyDataPermission, RevokePermission, SplitJoinPermission
from .roles import SimpleRole, ListRole, RoleLink
from .utils import HashSet


def __all_subclasses(cls):
    return set(cls.__subclasses__()).union([s for c in cls.__subclasses__() for s in __all_subclasses(c)])


def __load_class_map():
    for klass in __all_subclasses(Base):
        klass._set_api_type()
        java_class, _type = getattr(klass, 'JAVA_CLASS'), getattr(klass, 'API_TYPE')
        if java_class and _type:
            JAVA_CLASS_MAP[java_class] = klass
            TYPES_CLASS_MAP[_type] = klass


__load_class_map()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import warnings
from base64 import b64encode

from .base import Base
from .binder import Binder
from .crypto import KeyAddress, PrivateKey, PublicKey
from .hash_id import HashId
from .roles import Role
from .permissions import Permission
from .utils import HashSet
from universa.exceptions import UniversaException
from universa.utils import ut, dt


class Contract(Base):
    JAVA_CLASS = 'com.icodici.universa.contract.Contract'
    ROLES = ['creator', 'issuer', 'owner']

    def __init__(self, binary=None, **kwargs):
        self._binary = binary
        super(Contract, self).__init__(**kwargs)

    @classmethod
    def from_packed(cls, from_packed):
        raise NotImplementedError()

    def _instantiate_data(self):
        if self._binary is not None:
            return {'__type': 'binary', 'base64': self._binary}

    def __set_role(self, role_name, *addresses):
        """
        Set provided addresses to the role.
        It's recommended to use KeyAddress instances, not PublicKey.
        PrivateKey is not allowed!

        :type addresses: list[KeyAddress | str]
        :returns: Role
        """
        if role_name not in self.ROLES:
            return

        __addresses = []
        for address in addresses:
            __addr = None
            if isinstance(address, KeyAddress):
                __addr = address
            elif isinstance(address, PublicKey):
                warnings.warn('It\'s recommended to use KeyAddress instances, not PublicKey.',
                              PendingDeprecationWarning)
                __addr = address.short_address
            elif isinstance(address, PrivateKey):
                warnings.warn('PrivateKey instance is not allowed! Use KeyAddress instead.',
                              SyntaxWarning)
                __addr = address.public_key.short_address
            elif isinstance(address, str):
                try:
                    __addr = KeyAddress(address=address)
                except UniversaException as e:
                    if 'Not a Base58 input' in e.error.get('text', ''):
                        try:
                            __addr = KeyAddress(uaddress=address)
                        except UniversaException:
                            pass

            if __addr is not None:
                __addresses.append(__addr)

        cmd = 'set{}Keys'.format(role_name.capitalize())
        rsp = self._invoke(cmd, *__addresses)
        klass = self.get_class_by_java(rsp['className'])
        return klass.get(rsp['id'])

    def set_creator_keys(self, *keys):
        """Set 'creator' role to given keys."""
        warnings.warn('Contract.set_creator_keys(*keys) will be deprecated. '
                      'Use contract.set_creator_addresses(*addresses) instead.',
                      PendingDeprecationWarning)
        return self.set_creator_addresses(*keys)

    def set_issuer_keys(self, *keys):
        """Set 'issuer' role to given keys."""
        warnings.warn('Contract.set_issuer_keys(*keys) will be deprecated. '
                      'Use contract.set_issuer_addresses(*addresses) instead.',
                      PendingDeprecationWarning)
        return self.set_issuer_addresses(*keys)

    def set_owner_keys(self, *keys):
        """Set 'owner' role to given keys."""
        warnings.warn('Contract.set_owner_keys(*keys) will be deprecated. '
                      'Use contract.set_owner_addresses(*addresses) instead.',
                      PendingDeprecationWarning)
        return self.set_owner_addresses(*keys)

    def set_creator_addresses(self, *addresses):
        """Set 'creator' role to given addresses."""
        return self.__set_role('creator', *addresses)

    def set_issuer_addresses(self, *addresses):
        """Set 'issuer' role to given addresses."""
        return self.__set_role('issuer', *addresses)

    def set_owner_addresses(self, *addresses):
        """Set 'owner' role to given addresses."""
        return self.__set_role('owner', *addresses)

    def get_role(self, role_name):
        if role_name in self.ROLES:
            rsp = self._invoke('get{}'.format(role_name.capitalize()))
        else:
            rsp = self._invoke('getRole', role_name)

        if rsp is not None:
            klass = self.get_class_by_java(rsp['className'])
            return klass.get(rsp['id'])
        else:
            return None

    def get_creator(self):
        return self.get_role('creator')

    def get_issuer(self):
        return self.get_role('issuer')

    def get_owner(self):
        return self.get_role('owner')

    def create_role(self, name, role):
        """
        :type name: str
        :type role: Role
        :return: Role
        """
        rsp = self._invoke('createRole', name, role.remote)
        klass = self.get_class_by_java(rsp['className'])
        return klass.get(rsp['id'])

    def get_created_at(self):
        """
        :return: unix time
        :rtype: datetime | None
        """
        rsp = self._invoke('getCreatedAt')
        return dt(rsp['seconds'])
    created_at = property(get_created_at)

    def get_expires_at(self):
        """
        :return: unix time
        :rtype: datetime | None
        """
        rsp = self._invoke('getExpiresAt')
        return dt(rsp['seconds']) if rsp is not None else None

    def set_expires_at(self, expires_at):
        """
        :param expires_at: unix time
        :type expires_at: datetime
        """
        self._invoke('setExpiresAt', {'seconds': ut(expires_at), '__type': 'unixtime'} 
                                         if expires_at is not None 
                                         else None)
    expires_at = property(get_expires_at, set_expires_at)

    def add_permission(self, permission):
        """
        :type permission: Permission
        """
        self._invoke('addPermission', permission.remote)

    def get_permissions(self):
        raise NotImplementedError()
        rsp = self._invoke('getPermissions')

    def is_permitted(self, permission_name, *keys):
        """
        Checks if permission of given type that is allowed for given keys exists.

        :type permission_name: str
        :type keys: PublicKey
        :rtype: bool
        """
        return self._invoke('isPermitted', permission_name, [key.remote for key in keys])

    def check_applicable_permission_quantized(self, permission):
        """
        Quantize given permission (add cost for that permission).

        :type permission: Permission
        """
        self._invoke('checkApplicablePermissionQuantized', permission.remote)

    def seal(self):
        return self._invoke('seal')['base64']

    def check(self):
        return self._invoke('check')

    def get_errors(self):
        return self._invoke('getErrors')

    def get_packed_transaction(self):
        return self._invoke('getPackedTransaction')

    def get_origin(self):
        return HashId(id=self._invoke('getOrigin')['id'])

    def get_id(self):
        return HashId(id=self._invoke('getId')['id'])

    def get_state_data(self):
        rsp = self._invoke('getStateData')
        return Binder.get(rsp['id'])

    def set_state(self, name, value):
        remote = self._invoke_static_with_type('Binder', 'of', value)
        self._invoke('set', name,  remote)

    def get_processed_cost(self):
        return self._invoke('getProcessedCost')

    def get_processed_cost_u(self):
        return self._invoke('getProcessedCostU')

    def get_transaction_pack(self):
        rsp = self._invoke('getTransactionPack')
        return TransactionPack(id=rsp['id'])

    def get_new_items(self):
        return HashSet(id=self._invoke('getNewItems')['id'])

    def get_revoking_items(self):
        return HashSet(id=self._invoke('getRevokingItems')['id'])

    @classmethod
    def from_sealed_binary(cls, sealed_binary):
        """

        :type sealed_binary: bytes
        :rtype: Contract
        """
        rsp = cls._invoke_static('fromSealedBinary', {'__type': 'binary', 'base64': b64encode(sealed_binary).decode()})
        return Contract(id=rsp['id'])


class TransactionPack(Base):
    JAVA_CLASS = 'com.icodici.universa.contract.TransactionPack'

    def _instantiate_data(self):
        pass

    @classmethod
    def unpack(cls, pack_or_contract_bytes, allow_non_transactions=None):
        """
        :type pack_or_contract_bytes: bytes
        :type allow_non_transactions: bool | None
        """
        rsp = cls._invoke_static('unpack', {'__type': 'binary', 'base64': b64encode(pack_or_contract_bytes).decode()})
        return TransactionPack(id=rsp['id'])

    def set_contract(self, contract):
        """
        :type contract: Contract
        """
        self._invoke('setContract', contract)

    def get_contract(self):
        """
        :rtype: Contract
        """
        return Contract(id=self._invoke('getContract')['id'])


class ContractsService(Base):
    JAVA_CLASS = 'com.icodici.universa.contract.ContractsService'

    @classmethod
    def create_split_join(cls, contracts_to_join, amounts_to_split, addresses_to_split, owner_keys,
                          field_name='amount'):
        """

        :type contracts_to_join: list[Contract]
        :type amounts_to_split: list[str]
        :type addresses_to_split: list[KeyAddress]
        :type owner_keys: HashSet[PrivateKey]
        :type field_name: str
        :rtype: list[Contract]
        """
        return [Contract(id=row['id'])
                    for row in cls._invoke_static('createSplitJoin', contracts_to_join, amounts_to_split,
                                                  addresses_to_split, owner_keys, field_name)]


class Reference(Base):
    JAVA_CLASS = 'com.icodici.universa.contract.Reference'


class Parcel(Base):
    JAVA_CLASS = 'com.icodici.universa.contract.Parcel'

    def __init__(self, **kwargs):
        super(Parcel, self).__init__(**kwargs)

    def _instantiate_data(self):
        return

    @classmethod
    def of(cls, payload, u_contract, u_keys):
        """
        :type payload: Contract
        :type u_contract: Contract
        :type u_keys: list[PrivateKey]
        :rtype: Parcel
        """
        rsp = cls._invoke_static('of', payload, u_contract, u_keys)
        return Parcel(id=rsp['id'])

    def pack(self):
        return self._invoke('pack')

    def get_payload(self):
        return TransactionPack(id=self._invoke('getPayload')['id'])

    def get_payment(self):
        return TransactionPack(id=self._invoke('getPayment')['id'])


class ExtendedSignature(Base):
    JAVA_CLASS = 'com.icodici.universa.contract.ExtendedSignature'




# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import json

from .base import transport, Base


class HashSet(Base):
    JAVA_CLASS = 'java.util.HashSet'
    API_TYPE = 'Set'

    items = []

    def __init__(self, items=None, **kwargs):
        __items = []
        for item in items or []:
            if not isinstance(item, Base):
                klass = self.get_class_by_type(item['__type'])
                if klass.API_TYPE == 'KeyAddress':
                    item = klass(uaddress=item['uaddress']['base64'])

            if item is not None:
                __items.append(item)

        self.__items = __items
        super(HashSet, self).__init__(**kwargs)
        self.__position = 0

    def __iter__(self):
        for obj in self._invoke('toArray'):
            klass = self.get_class_by_java(obj['className'])
            yield klass.get(obj['id'])

    def __len__(self):
        return self._invoke('size')

    def __getitem__(self, item):
        obj = self._invoke('toArray')[item]
        klass = self.get_class_by_java(obj['className'])
        return klass.get(obj['id'])

    @classmethod
    def get(cls, _id):
        instance = cls(id=_id, items=cls._get(_id), skip_update=True)
        transport.OBJECTS[_id] = instance
        return instance

    def _instantiate_data(self):
        items = [item.remote if isinstance(item, Base) else item for item in self.__items]
        del self.__items
        return items



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import gc
import logging
import os
import subprocess
import time
import unittest

try:
    _FileNotFoundError = FileNotFoundError
except NameError:
    _FileNotFoundError = OSError

from streamexpect import ExpectTimeout

from universa.transport import transport
from universa.types import PrivateKey


logger = logging.getLogger()


class TestTransport(unittest.TestCase):
    TCP_SOCKET_HOST = '127.0.0.1'
    TCP_SOCKET_PORT = 59300
    UNIX_SOCKET_PATH = '/tmp/umi-socket-for-tests'

    umi_tcp_server = None
    umi_unix_server = None

    @classmethod
    def setUpClass(cls):
        try:
            os.remove(cls.UNIX_SOCKET_PATH)
        except _FileNotFoundError:
            pass

        cls.umi_tcp_server = subprocess.Popen(['umi', '--noexit', '--listen', 'tcp://{}:{}'.format(cls.TCP_SOCKET_HOST, cls.TCP_SOCKET_PORT)])
        cls.umi_unix_server = subprocess.Popen(['umi', '--noexit', '--listen', 'unix://{}'.format(cls.UNIX_SOCKET_PATH)])
        time.sleep(3)  # Wait a bit while both UMI servers start

    @classmethod
    def tearDownClass(cls):
        for attr in ('umi_tcp_server', 'umi_unix_server'):
            proc = getattr(cls, attr, None)
            if proc is not None:
                proc.terminate()

    def test_setupUMI(self):
        with self.assertRaises(ValueError):
            transport.setupUMI('pipe', 'veryrandombinaryname4242')
        transport.setupUMI('pipe', transport.DEFAULT_BINARY)

        # Test TCP and UNIX sockets
        i, config = 0, ('', {})
        configs = [
            ('tcp', {'host': self.TCP_SOCKET_HOST, 'port': self.TCP_SOCKET_PORT}),
            ('unix', {'path': self.UNIX_SOCKET_PATH}),
            ('pipe', {'path': transport.DEFAULT_BINARY}),
            ('tcp', {'host': self.TCP_SOCKET_HOST, 'port': self.TCP_SOCKET_PORT}),
        ]

        try:
            for i, config in enumerate(configs):
                method, connection_args = config
                transport.setupUMI(method, **connection_args)
                test = transport.test()
                self.assertEqual(test['result']['system'], 'UMI', 'connection to UMI failed')
                self.assertEqual(test['ref'], 0, 'invalid ref')
                time.sleep(1)
        except ExpectTimeout as e:
            logger.exception('UMI request failed with config %s: %s' % (i, str(config)))
            raise e

    def test_transport_GC(self):
        gc.enable()

        transport.setupUMI('pipe', transport.DEFAULT_BINARY)

        self.assertEqual(len(transport.OBJECTS), 0, 'transport objects are not blank')

        key = PrivateKey(size=2048)
        self.assertEqual(len(transport.OBJECTS), 1, 'transport objects does not contains only one value')

        key2 = PrivateKey(size=2048)
        self.assertEqual(len(transport.OBJECTS), 2, 'transport objects does not contains only two values')

        # Create a key overriding previous one variable.
        # GarbageCollector should delete old object and kill it from transport (and UMI remote object).
        key2 = PrivateKey(size=2048)
        self.assertEqual(len(transport.OBJECTS), 2, 'transport objects does not contains only two values')

        # Create a key and don't save it to the variable. It should be deleted by UMI at once.
        PrivateKey(size=2048)
        self.assertEqual(len(transport.OBJECTS), 2, 'transport objects does not contains only two values')

        # Delete a local key object. It should be removed from UMI too.
        del key2
        self.assertEqual(len(transport.OBJECTS), 1, 'transport objects does not contains only one value')


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

from universa import logging
from universa.exceptions import UniversaException
from universa.transport import transport


logger = logging.getLogger()


class Base(object):
    JAVA_CLASS = ''
    API_TYPE = ''

    LOCAL_ATTRS = ('id',)

    __slots__ = ('id', '__weakref__')

    def __init__(self, **kwargs):
        self.id = kwargs.get('id')

        if self.id is None:
            for k, v in self.from_instantiate(self._instantiate()).items():
                setattr(self, k, v)
            transport.OBJECTS[self.id] = self

    def __str__(self):
        return '<{}: {}>'.format(self.__class__.__name__, self.id)

    def __repr__(self):
        return '<{}: {}>'.format(self.__class__.__name__, self.id)

    def __del__(self):
        logger.debug('Deleting local %s', self)
        if self.id and transport:
            transport.drop_objects([self.id])

    def __getattr__(self, item):
        if item in self.LOCAL_ATTRS:
            return self.__getattribute__(item)

        if self.id is not None:
            try:
                data = self._get_field(self.id, item)
            except UniversaException:
                raise AttributeError('{class_name!r} object has no attribute {attr!r}'
                                         .format(class_name=self.__class__.__name__, attr=item))
            return data

    # def __setattr__(self, key, value):
    #     if key in self.LOCAL_ATTRS:
    #         super(Base, self).__setattr__(key, value)
    #
    #     if self.id is not None:
    #         self._invoke('set', key, value)
    #     else:
    #         raise AttributeError

    def __eq__(self, other):
        return self.equals(other)

    def __ne__(self, other):
        return not self.equals(other)

    @classmethod
    def _set_api_type(cls):
        cls.API_TYPE = cls.API_TYPE or cls.JAVA_CLASS.split('.')[-1].replace('$', '.')

    @staticmethod
    def get_class_by_java(java_class_name):
        return JAVA_CLASS_MAP.get(java_class_name, DEFAULT_CLASS)

    @staticmethod
    def get_class_by_type(api_type):
        return TYPES_CLASS_MAP.get(api_type, DEFAULT_CLASS)

    @classmethod
    def obj_umification(cls, obj):
        if isinstance(obj, Base):
            obj = obj.remote
        elif isinstance(obj, dict):
            obj = cls.kwargs_umification(obj)
        elif isinstance(obj, (tuple, list, set)):
            obj = cls.args_umification(obj)
        return obj

    @classmethod
    def args_umification(cls, args):
        umi_args = []
        for obj in args:
            umi_args.append(cls.obj_umification(obj))
        return umi_args

    @classmethod
    def kwargs_umification(cls, dct):
        umi_dct = {}
        for key, value in dct.items():
            umi_dct[cls.obj_umification(key)] = cls.obj_umification(value)
        return umi_dct

    def _instantiate(self):
        args = [self.__class__.API_TYPE]
        data = self._instantiate_data()
        if data is not None:
            if isinstance(data, tuple):
                args.extend(data)
            else:
                args.append(data)
        return transport.instantiate(*args)

    def _invoke(self, method_name, *args):
        return transport.invoke(self.id, method_name, *self.args_umification(args))

    @classmethod
    def _invoke_static_with_type(cls, api_type, method_name, *args):
        return transport.invoke_static(api_type, method_name, *cls.args_umification(args))

    @classmethod
    def _invoke_static(cls, method_name, *args):
        return cls._invoke_static_with_type(cls.API_TYPE, method_name, *cls.args_umification(args))

    @classmethod
    def _get(cls, _id):
        return cls.from_get(transport.get(_id))

    @classmethod
    def _get_field(cls, _id, field_name):
        return cls.from_get(transport.get_field(_id, field_name))

    @classmethod
    def get(cls, _id):
        data = cls._get(_id)
        instance = cls(id=_id, skip_update=True, **data)
        # TODO: probably we don't want to save this object to the transport cause it may not be saved to a variable
        transport.OBJECTS[_id] = instance
        return instance

    def update(self):
        for k, v in self._get(self.id).items():
            setattr(self, k, v)

    @property
    def remote(self):
        if isinstance(self, Base) and self.id is not None:
            return {'__type': 'RemoteObject', 'className': self.JAVA_CLASS, 'id': self.id}

    def _get_hashset(self, cmd):
        rsp = self._invoke(cmd)
        if rsp is not None:
            klass = self.get_class_by_java(rsp['className'])
            return klass.get(rsp['id'])
        else:
            return None

    def _equals(self, equals_method_name, obj):
        if id(self) == id(obj):
            return True

        if self.JAVA_CLASS != obj.JAVA_CLASS:
            return False

        return self._invoke(equals_method_name, obj.remote)

    def equals(self, obj):
        return self._equals('equals', obj)

    def _instantiate_data(self):
        raise NotImplementedError()

    @staticmethod
    def from_instantiate(data):
        return {'id': data['id']}

    @staticmethod
    def from_get(data):
        return data


DEFAULT_CLASS = Base
TYPES_CLASS_MAP = {}
JAVA_CLASS_MAP = {}



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

from base64 import b64encode

from .base import logger, Base


#
# Abstract
#
class AbstractKey(Base):
    def __init__(self, size=None, key=None, **kwargs):
        self.size = size
        self.key = key
        super(AbstractKey, self).__init__(size=size, key=key, **kwargs)
        if self.id is not None and not kwargs.get('skip_update', False):
            self.update()
        self.pack()

    def _instantiate_data(self):
        if self.key is not None:
            return self.to_binary()
        elif self.size is not None:
            return self.size
        else:
            logger.exception('Can\'t find .instantiate_data() data for %s', self)
            return None

    def to_binary(self):
        return {'__type': 'binary', 'base64': self.key}

    def pack(self):
        rsp = self._invoke('pack')
        self.key = rsp['base64']
        return self.key


#
# Real
#
class KeyAddress(Base):
    JAVA_CLASS = 'com.icodici.crypto.KeyAddress'

    # __slots__ = ('id', 'address', 'uaddress')

    def __init__(self, address=None, uaddress=None, **kwargs):
        self.address = address
        self.uaddress = uaddress
        super(KeyAddress, self).__init__(address=address, uaddress=uaddress, **kwargs)
        if self.address is None:
            self.to_string()
        if self.uaddress is None:
            self.update()

    def _instantiate_data(self):
        if self.address:
            return self.address
        elif self.uaddress:
            return self.to_binary()

    def to_binary(self):
        return {'__type': 'binary', 'base64': self.uaddress}

    @staticmethod
    def from_get(data):
        return {'uaddress': data['uaddress']['base64']}

    def is_matching_key_address(self, key_address):
        if key_address is None or self.JAVA_CLASS != key_address.JAVA_CLASS or key_address.id is None:
            return False

        rsp = self._invoke('isMatchingKeyAddress', key_address.remote)
        return rsp

    def get_packed(self):
        if self.uaddress:
            return self.uaddress
        self.uaddress = self._invoke('getPacked')['base64']
        return self.uaddress

    @property
    def is_long(self):
        return self._invoke('isLong')

    def to_string(self):
        self.address = self._invoke('toString')
        return self.address


class PrivateKey(AbstractKey):
    JAVA_CLASS = 'com.icodici.crypto.PrivateKey'

    # __slots__ = ('id', 'size', 'packed', 'key')

    def __init__(self, size=None, key=None, **kwargs):
        if isinstance(key, bytes):
            key = b64encode(key).decode()
        super(PrivateKey, self).__init__(size=size, key=key, **kwargs)
        self._public_key = None

    @staticmethod
    def from_get(data):
        return {'packed': data['packed']}

    def get_public_key(self):
        rsp = self._invoke('getPublicKey')
        return PublicKey.get(rsp['id'])

    @property
    def public_key(self):
        """
        :rtype: PublicKey
        """
        return self.get_public_key()


class PublicKey(AbstractKey):
    JAVA_CLASS = 'com.icodici.crypto.PublicKey'

    # __slots__ = ('id', 'size', 'key')

    def __init__(self, **kwargs):
        self._short_address, self._long_address = None, None
        super(PublicKey, self).__init__(**kwargs)

    @staticmethod
    def from_get(data):
        return {'key': data['packed']['base64']}

    def _get_address(self, short=True):
        """
        :rtype: KeyAddress
        """
        rsp = self._invoke('getShortAddress' if short else 'getLongAddress')
        return KeyAddress.get(rsp['id'])

    @property
    def short_address(self):
        """
        :rtype: KeyAddress
        """
        return self.get_short_address()

    def get_short_address(self):
        """
        :rtype: KeyAddress
        """
        return self._get_address(short=True)

    @property
    def long_address(self):
        """
        :rtype: KeyAddress
        """
        return self.get_long_address()

    def get_long_address(self):
        """
        :rtype: KeyAddress
        """
        return self._get_address(short=False)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import json

from .base import Base


class Permission(Base):
    def __init__(self, name=None, role=None, params=None, **kwargs):
        """

        :type name: str | None
        :type role: universa.types.roles.Role | None
        :type params: dict | None
        """
        self._name = name
        self._role = role
        self._params = params
        super(Permission, self).__init__(**kwargs)

    def _instantiate_data(self):
        if all(i is not None for i in [self._name, self._role, self._params]):
            return self._name, self._role.remote, self._params
        elif self._name is not None and self._role is not None:
            return self._name, self._role.remote
        else:
            return None

    def compare_to(self, obj):
        return self._equals('compareTo', obj)

    def get_id(self):
        """
        :rtype: int | None
        """
        return self._invoke('getId')

    def set_id(self, _id):
        """
        :type _id: int
        """
        return self._invoke('setId', _id)

    def get_name(self):
        """
        :rtype: str
        """
        return self._invoke('getName')

    def get_params(self):
        """
        :rtype:
        """
        raise NotImplementedError()
        return self._invoke('getParams')

    def get_role(self):
        """
        :rtype: universa.types.roles.Role | None
        """
        rsp = self._invoke('getRole')
        if rsp is not None:
            klass = self.get_class_by_java(rsp['className'])
            return klass.get(rsp['id'])
        else:
            return None

    def is_allowed_for_keys(self, *keys):
        """
        :type keys: universa.types.crypto.PublicKey
        :rtype: bool
        """
        return self._invoke('isAllowedForKeys', *[key.remote for key in keys])

    # TODO:
    # abstract void	checkChanges(Contract contract, Contract changed, java.util.Map<java.lang.String,net.sergeych.diff.Delta> stateChanges, java.util.Set<Contract> revokingItems, java.util.Collection<com.icodici.crypto.PublicKey> keys)
    # void	deserialize(net.sergeych.tools.Binder data, net.sergeych.biserializer.BiDeserializer deserializer)
    # static Permission	forName(java.lang.String name, Role role, net.sergeych.tools.Binder params)
    # net.sergeych.tools.Binder	getParams()
    # net.sergeych.tools.Binder	serialize(net.sergeych.biserializer.BiSerializer serializer)



class ChangeNumberPermission(Permission):
    JAVA_CLASS = 'com.icodici.universa.contract.permissions.ChangeNumberPermission'


class ChangeOwnerPermission(Permission):
    JAVA_CLASS = 'com.icodici.universa.contract.permissions.ChangeOwnerPermission'


class ModifyDataPermission(Permission):
    JAVA_CLASS = 'com.icodici.universa.contract.permissions.ModifyDataPermission'


class RevokePermission(Permission):
    JAVA_CLASS = 'com.icodici.universa.contract.permissions.RevokePermission'

    def __init__(self, role, **kwargs):
        """
        :type role: universa.types.roles.Role
        """
        super(RevokePermission, self).__init__(role=role, **kwargs)

    def _instantiate_data(self):
        return self._role.remote


class SplitJoinPermission(Permission):
    JAVA_CLASS = 'com.icodici.universa.contract.permissions.SplitJoinPermission'



import setuptools


CLASSIFIERS = [
    'Development Status :: 3 - Alpha',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'License :: OSI Approved :: MIT License',
    'Operating System :: OS Independent',
    'Intended Audience :: Developers',
    'Topic :: System :: Networking',
]


with open('README.md', 'r') as fh:
    long_description = fh.read()


EXTRAS_REQUIRE = {
    'dev': [
        'pytest',
    ]
}


setuptools.setup(
    name='universa',
    version='0.2.0',
    author='Vadim Kovrigin',
    author_email='kovrigin.dev@gmail.com',
    description='Python package to access Universa API from python.',
    include_package_data=True,
    install_requires=[
        'pexpect>=4.7.0',
        'streamexpect>=0.2.1',
    ],
    extras_require=EXTRAS_REQUIRE,
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords=['Universa'],
    url='https://github.com/vkovrigin/universa',
    packages=setuptools.find_packages(),
    classifiers=CLASSIFIERS,
)
